# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-06-09 09:26
from __future__ import unicode_literals

import chemtrails.contrib.permissions.forms.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0005_guild'),
    ]

    operations = [
        migrations.CreateModel(
            name='JSONModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('json', chemtrails.contrib.permissions.forms.fields.JSONField(default=dict)),
            ],
        ),
    ]
