# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-10-04 23:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('chemtrails_permissions', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='accessrule',
            name='direction',
            field=models.SmallIntegerField(blank=True, choices=[(-1, 'INCOMING'), (0, 'EITHER'), (1, 'OUTGOING')], default=None, help_text='Override calculated direction for relationships.<br/> Use with caution, this will affect *all* directional relationships for evaluation of this rule!', null=True, verbose_name='direction'),
        ),
    ]
